// Setup.cpp 
//

#include "stdafx.h"
#include "Setup.h"
#include "dialogs.h"
#include "winreg.h"
#include "winver.h"
#include "winnetwk.h"
#include "objbase.h"
#include "shlobj.h"
#include "winnls.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

int MySscanf(LPCSTR buf, LPCSTR Format, ...);
LPCSTR StripPath(LPCSTR Path);
HRESULT CreateLink(LPCSTR lpszPathObj, LPCSTR lpszPathLink, LPCSTR lpszDesc);
BOOL IncRegCnt(LPCSTR Path);
BOOL GetFilePermissions(LPCSTR Path, DWORD& Perms);

extern BOOL DDECreateGroup (LPCSTR GroupName);
extern BOOL DDEAddProgramItem(LPCSTR szProgName, LPCSTR szExePath);
extern BOOL DDEGroupRetrieval (CComboBox& ListBox);
extern BOOL DDEIsFolderEmpty(LPCSTR FolderName);

FILE *CSetupApp::m_LogFp;

OSVERSIONINFO OsInfo;

// Some usful Paths...
char WinDir[MAX_PATH];
char SystemDir[MAX_PATH];
char ProgramFolderDir[MAX_PATH];
char UserProgramFolderDir[MAX_PATH];
char ProgramDir[MAX_PATH];
char DestPath[MAX_PATH];
char FolderName[MAX_PATH];
int  OsType;

int MySscanf(LPCSTR buf, LPCSTR Format, ...)
{
	va_list marker;
	va_start(marker, Format);
	int Rval = 0;
	char *p;
	int *i;
	int base;
	
	while(*Format)
	{
		if (*Format == '%')
		{
			switch(*++Format)
			{
				case 's':
					p = va_arg(marker,char *);
					while(*buf && (!Format[1] || (*buf != Format[1])))
						*p++ = *buf++;
					*p = 0;
					break;
				case 'l':
					++Format; // %ld, skip over one to get in sync
				case 'O':
					base = 10;
					i = va_arg(marker, int *);
					*i = 0;
					if (*buf == '0')
						base = 8; // Octal
					while(*buf && isdigit(*buf))
					{
						*i *= base;
						*i += *buf - '0';
						++buf;
					}
					break;
				case 'd':
					base = 10;
					i = va_arg(marker, int *);
					*i = 0;
					while(*buf && isdigit(*buf))
					{
						*i *= base;
						*i += *buf - '0';
						++buf;
					}
					break;
			}
			++Format;
		}
		else
		{
			while(*buf && (*buf != *Format))
				++buf;
			++Format;
			++buf;
		}
	}

	va_end(marker);
	return Rval;
}


/////////////////////////////////////////////////////////////////////////////
// CSetupApp

BEGIN_MESSAGE_MAP(CSetupApp, CWinApp)
	//{{AFX_MSG_MAP(CSetupApp)
		// NOTE - the ClassWizard will add and remove mapping macros here.
		//    DO NOT EDIT what you see in these blocks of generated code!
	//}}AFX_MSG
	ON_COMMAND(ID_HELP, CWinApp::OnHelp)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSetupApp construction

CSetupApp::CSetupApp()
{
	// CG: The following line was added by the OLE Automation component.
	EnableAutomation();
	m_FileClassRoot = NULL;
	m_DoReboot = FALSE;
}

/////////////////////////////////////////////////////////////////////////////
// The one and only CSetupApp object

CSetupApp theApp;

/////////////////////////////////////////////////////////////////////////////
// CSetupApp initialization

BOOL CSetupApp::InitInstance()
{
	// Set up a initializig message
	CInitDlg *Init = new CInitDlg;
	CInstallSheet InstallTab;

	Init->Create(IDD_INIT);
	Init->ShowWindow(SW_SHOW);
	CoInitialize(NULL);

	// Check OS and environment
	OsInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&OsInfo);
	if (OsInfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
	{
		OsType = OS_WIN95;
	}
	else if (OsInfo.dwPlatformId == VER_PLATFORM_WIN32_NT)
	{
		if (OsInfo.dwMajorVersion >= 4)
			OsType = OS_NT4;
		else
			OsType = OS_NT351; // Put some more checking here in the future
	}

	if (OsType == OS_UNKNOWN)
	{
		AfxMessageBox("Unknown version of Windows. Install aborted.");
		return FALSE;
	}

	// Initialize usful globals...
	switch(OsType)
	{
		case OS_WIN95: if (!LoadWin95Info()) return FALSE; break;
		case OS_NT4: if (!LoadNT4Info()) return FALSE;	break;
		case OS_NT351: if (!LoadNt351Info()) return FALSE; break;
	}

	// Load the files we shall transfere, based on the current OS
	if (!LoadInfFile())
		return FALSE;

	// Remove the initialize window and set up the Install tab...
	delete Init;
	
	if (InstallTab.DoModal() == ID_WIZFINISH)
	{
		// Set up a progress dialog
		CProgress *Progress = new CProgress;
		int NumFiles = 0;
		CFileList *fl;
		CClassList *cl;
		int Index;
		CString cBuf, cTmp;

		Progress->Create(IDD_PROGRESS);
		Progress->ShowWindow(SW_SHOW);
		Progress->SetFocus();

		Progress->m_Step.SetWindowText("Prepering to copy files");
		strcpy(DestPath,InstallTab.m_Tab1.m_DstPath);
		strcpy(FolderName,InstallTab.m_Tab4.m_Folder);

		// Count the files we will copy
		for(cl = m_FileClassRoot, Index = 0; cl; cl = cl->m_Next, Index++)
		{
			switch(InstallTab.m_Tab2.m_Mode)
			{
				case 0: // Normal
					if (!cl->m_IsTypical)
					{
						cl->m_Include = FALSE;
						continue;
					}
					cl->m_Include = TRUE;
					break;

				case 1: // Minimum
					if (!cl->m_IsMinimum)
					{
						cl->m_Include = FALSE;
						continue;
					}
					cl->m_Include = TRUE;
					break;

				case 2: // Custom
					if (Index < 10)
					{
						if (!InstallTab.m_Tab3.m_Module[Index])
						{
							cl->m_Include = FALSE;
							continue;
						}
					}
					cl->m_Include = TRUE;
					break;
			}

			for(fl = cl->m_FileList; fl; fl = fl->m_Next)
			{
				++NumFiles;				
			}
		}

		// We are ready to start copying files.
		// Open the log file that Uninstall will use
		if (!CreateDirectory(InstallTab.m_Tab1.m_DstPath,NULL))
		{
			if (GetLastError() != ERROR_ALREADY_EXISTS)
			{
				AfxMessageBox("Failed to create destination directory. Install aborted.");
				return FALSE;
			}
		}

		cBuf.Format("%s\\.Install.Log",
			InstallTab.m_Tab1.m_DstPath);

		if ((m_LogFp = fopen(cBuf,"w")) == NULL)
		{
			AfxMessageBox("Failed to create Log file. Install aborted.");
			return FALSE;
		}

		cBuf.Format("%s\\.UnInst.inf",
			InstallTab.m_Tab1.m_DstPath);
	
		// Open uninstall log file. If the file exist, we assume upgrade and
		// append to the end of it. That way uninstall should work also after
		// an upgrade.
		if ((m_UnInstFp = fopen(cBuf,"a+")) == NULL)
		{
			AfxMessageBox("Failed to create Uninstall information file. Install aborted.");
			return FALSE;
		}

		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"Ready to install %s to %s\n", m_Name, InstallTab.m_Tab1.m_DstPath);

		fseek(m_UnInstFp,0L,SEEK_END);
		if (!ftell(m_UnInstFp))
		{
			fprintf(m_UnInstFp,"%s\n", (LPCSTR)m_Name);
			fprintf(m_UnInstFp,"dir %s\n", InstallTab.m_Tab1.m_DstPath);
		}

		// Start the copy process
		Progress->m_Progress.SetRange(0,NumFiles);
		Progress->m_Progress.SetStep(1);

		for(cl = m_FileClassRoot, Index = 0; cl; cl = cl->m_Next, Index++)
		{
			if (!cl->m_Include)
			{
				m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"Skipping %s...\n", cl->m_Name);
				continue;
			}

			cBuf.Format("Copying files (%s)...", cl->m_Name);
			Progress->m_Step.SetWindowText(cBuf);

			for(fl = cl->m_FileList; fl; fl = fl->m_Next)
			{
				Progress->m_Current.SetWindowText(fl->m_DstName);
				InstallFile(cl,fl);
				Progress->m_Progress.StepIt();
			}
		}

		Progress->m_Current.SetWindowText("");
		Progress->m_Step.SetWindowText("Registering Uninstall program...");


		cBuf.Format("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\%s",
			m_Name);

		CreateRegisteryKey(HKEY_LOCAL_MACHINE,cBuf);

		cBuf.Format("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\%s\\DisplayName",
			m_Name);
		SetRegisteryString(HKEY_LOCAL_MACHINE,cBuf,m_Name,m_Name.GetLength());

		cBuf.Format("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\%s\\UninstallString",
			m_Name);
		cTmp.Format("%s\\Uninstall.exe \"%s\\.UnInst.inf\"", DestPath, DestPath);
		SetRegisteryString(HKEY_LOCAL_MACHINE,cBuf,cTmp,cTmp.GetLength());

		delete Progress;
		fclose(m_UnInstFp);
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\nInstall done.%s\n", m_DoReboot ? " Machine must reboot" : "");
		fclose(m_LogFp);

		if (m_DoReboot)
			AfxMessageBox("Installation complete. You must now reboot the machine.");
		else
			AfxMessageBox("Installation complete. You can now run the installed program.");
	}

	return FALSE;
}


BOOL CSetupApp::LoadNT4Info()
{
	if (!GetRegisteryString(
		HKEY_LOCAL_MACHINE,
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\ProgramFilesDir",
		ProgramDir,
		sizeof(ProgramDir)))
	{
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"**Failed to get NT4 Program dir\n");
		AfxMessageBox("Failed to get registery value for ProgramDir");
		*ProgramDir = 0;
	}

	if (!GetRegisteryString(
		HKEY_LOCAL_MACHINE,
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders\\Common Programs",
		ProgramFolderDir,
		sizeof(ProgramFolderDir)))
	{
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"**Failed to get NT4 ProgramFolder dir\n");
		AfxMessageBox("Failed to get registery value for ProgramFolderDir");
		*ProgramFolderDir = 0;
	}

	if (!GetRegisteryString(
		HKEY_CURRENT_USER,
		"Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\User Shell Folders\\Programs",
		UserProgramFolderDir,
		sizeof(UserProgramFolderDir)))
	{
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"**Failed to get NT4 UserProgramFolder dir\n");
		AfxMessageBox("Failed to get registery value for UserProgramFolderDir");
		*UserProgramFolderDir = 0;
	}

	GetWindowsDirectory(WinDir,sizeof(WinDir));
	GetSystemDirectory(SystemDir,sizeof(SystemDir));

	return TRUE;
}

BOOL CSetupApp::LoadWin95Info()
{
	if (!GetRegisteryString(
		HKEY_LOCAL_MACHINE,
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\ProgramFilesDir",
		ProgramDir,
		sizeof(ProgramDir)))
	{
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"**Failed to get Win95 Program dir\n");
		AfxMessageBox("Failed to get registery value for ProgramDir");
		*ProgramDir = 0;
	}

	if (!GetRegisteryString(
		HKEY_CURRENT_USER,
		"Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders\\Programs",
		UserProgramFolderDir,
		sizeof(UserProgramFolderDir)))
	{
		DWORD err = GetLastError();
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"**Failed to get Win95 UserProgramFolder dir\n");
		AfxMessageBox("Failed to get registery value for UserProgramFolderDir");
		*UserProgramFolderDir = 0;
	}

	
	if (!GetRegisteryString(
		HKEY_CURRENT_USER,
		"Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders\\Programs",
		ProgramFolderDir,
		sizeof(ProgramFolderDir)))
	{
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"**Failed to get Win95 ProgramFolder dir\n");
		AfxMessageBox("Failed to get registery value for ProgramFolderDir");
		*ProgramFolderDir = 0;
	}

	GetWindowsDirectory(WinDir,sizeof(WinDir));
	GetSystemDirectory(SystemDir,sizeof(SystemDir));

	return TRUE;
}


BOOL CSetupApp::LoadNt351Info()
{
	GetWindowsDirectory(WinDir,sizeof(WinDir));
	GetSystemDirectory(SystemDir,sizeof(SystemDir));
	sprintf(ProgramDir,"%c:", *WinDir);
	return TRUE;
}


BOOL CSetupApp::GetRegisteryString(HKEY MainKey, LPCSTR Path, LPSTR Buf, size_t Buflen)
{
	DWORD dwNihl, dwSize = Buflen;
	LONG Rval;
	HKEY hkey;
	char szKey[MAX_PATH];
	LPSTR Tag;

	m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"Trying to get registery key: %s ", Path);

	strcpy(szKey, Path);
	Tag = strrchr(szKey,'\\');
	ASSERT(Tag && (*Tag == '\\'));
	*Tag = 0;

	Tag = strrchr(Path,'\\');
	++Tag;

	if ((Rval = RegOpenKeyEx(
		MainKey,
		szKey,
		0,
		KEY_QUERY_VALUE,
		&hkey)) == ERROR_SUCCESS) 
	{
		if ((Rval = RegQueryValueEx(
			hkey,
			Tag,
			NULL,
			&dwNihl,
			(LPBYTE)Buf,
			&dwSize)) == ERROR_SUCCESS)
		{
			if (dwNihl == REG_EXPAND_SZ)
			{
				char *MyBuf = new char[Buflen];
				if (ExpandEnvironmentStrings(Buf,MyBuf,Buflen))
					strcpy(Buf,MyBuf);
				delete MyBuf;
			}
			RegCloseKey(hkey);
			m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"OK\n");
			return TRUE;
		}
		RegCloseKey(hkey);
	}

	m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"**Failed\n");
	return FALSE;
}

BOOL CSetupApp::SetRegisteryString(HKEY MainKey, LPCSTR Key, LPCSTR Buf, size_t Buflen)
{
	DWORD dwSize = Buflen;
	LONG Rval;
	HKEY hkey;
	char szKey[MAX_PATH];
	LPSTR Tag;

	m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"Trying to set registery key: %s ", Key);

	strcpy(szKey, Key);
	Tag = strrchr(szKey,'\\');
	ASSERT(Tag && (*Tag == '\\'));
	*Tag = 0;

	Tag = strrchr(Key,'\\');
	++Tag;

	if (RegOpenKeyEx(
		MainKey,
		szKey,
		0,
		KEY_WRITE,
		&hkey) == ERROR_SUCCESS) 
	{
		if ((Rval = RegSetValueEx(
			hkey,
			Tag,
			0,
			REG_SZ,
			(LPBYTE)Buf,
			dwSize)) == ERROR_SUCCESS)
		{
			fprintf(m_UnInstFp,"key %s %s\n", 
				(MainKey == HKEY_CLASSES_ROOT) ? "HKEY_CLASSES_ROOT"
				: (MainKey == HKEY_CURRENT_USER) ? "HKEY_CURRENT_USER"
				: (MainKey == HKEY_LOCAL_MACHINE) ? "HKEY_LOCAL_MACHINE" : "HKEY_USERS",
				Key);
			m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"OK\n");
			RegCloseKey(hkey);
			return TRUE;
		}
		RegCloseKey(hkey);
	}

	m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"**Failed\n");
	return FALSE;
}


BOOL CSetupApp::CreateRegisteryKey(HKEY MainKey, LPCSTR Key)
{
	DWORD dwNihl;
	LONG Rval;
	HKEY hkey;
	

	m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"Trying to create registery key: %s ", Key);

	if ((Rval = RegCreateKeyEx(
		MainKey,
		Key,
		0,
		"",
		REG_OPTION_NON_VOLATILE,
		KEY_SET_VALUE,
		NULL,
		&hkey,
		&dwNihl)) == ERROR_SUCCESS)
	{
		RegCloseKey(hkey);
		fprintf(m_UnInstFp,"key %s %s\n", 
			(MainKey == HKEY_CLASSES_ROOT) ? "HKEY_CLASSES_ROOT"
			: (MainKey == HKEY_CURRENT_USER) ? "HKEY_CURRENT_USER"
			: (MainKey == HKEY_LOCAL_MACHINE) ? "HKEY_LOCAL_MACHINE" : "HKEY_USERS",
			Key);
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"OK\n");
		return TRUE;
	}
	m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"**Failed\n");
	return FALSE;
}

BOOL CSetupApp::LoadInfFile()
{
	FILE *fp;
	char ReadBuf[MAX_PATH * 3];
	char buf[MAX_PATH * 2];
	int Version = 0;
	CClassList *cl;
	CFileList *fl;
	BOOL IsScanningClass = FALSE;
	int Index = LB_ERR;
	int Ofs;

	if ((fp = fopen("Setup.inf","r")) == NULL)
	{
		AfxMessageBox("Failed to open file.");
		return FALSE;
	}

	// Read header line
	fgets(ReadBuf,sizeof(ReadBuf),fp);
	MySscanf(ReadBuf,"\1%s\1 %d", buf, &Version);

	if (Version < 1)
	{
		AfxMessageBox("Unknown .inf file format.");
		fclose(fp);
		return FALSE;
	}

	fgets(ReadBuf,sizeof(ReadBuf),fp);
	m_Name = ReadBuf;
	m_Name.TrimRight();

	fgets(ReadBuf,sizeof(ReadBuf),fp);
	m_Menu = ReadBuf;
	m_Menu.TrimRight();

	if (Version > 1)
	{
		fgets(ReadBuf,sizeof(ReadBuf),fp);
		m_InstallDir = ReadBuf;
		m_InstallDir.TrimRight();
	}

	fgets(ReadBuf,sizeof(ReadBuf),fp);
	MySscanf(ReadBuf,"%d %d %d", &m_NT351, &m_NT4, &m_Win95);

	switch(OsType)
	{
		case OS_WIN95:
			if (!m_Win95)
			{
				AfxMessageBox("This program can not be installed on Windows 95");
				return FALSE;
			}
			break;
		case OS_NT4:
			if (!m_NT4)
			{
				AfxMessageBox("This program can not be installed on NT 4");
				return FALSE;
			}
			break;
		case OS_NT351: 
			if (!m_NT351)
			{
				AfxMessageBox("This program can not be installed on NT 3.51");
				return FALSE;
			}
			break;
	}	

	// Scan for file classes
	while(fgets(ReadBuf,sizeof(ReadBuf),fp) != NULL)
	{
		if (*ReadBuf == '@')
		{
			Ofs = 1;
			if (ReadBuf[1] == '@')
				Ofs = 2;

			if (ReadBuf[Ofs] == 1)
			{
				IsScanningClass = TRUE;
				// Accept this class
				cl = new CClassList;
				cl->m_Next = NULL;
				cl->m_FileList = NULL;
				
				MySscanf(ReadBuf +Ofs,"\1%s\1 %d %d %d %d %d %d %d \1%s\1\n",
				cl->m_Name,
				&cl->m_DoNT351,
				&cl->m_DoNT4,
				&cl->m_DoWin95,
				&cl->m_IsShared,
				&cl->m_IsSystem,
				&cl->m_IsTypical,
				&cl->m_IsMinimum,
				&cl->m_InstallDir);

				switch(OsType)
				{
					case OS_WIN95:
						if (!cl->m_DoWin95)
						{
skip:					
							delete cl;
							IsScanningClass = FALSE;
							continue;
						}
						break;
					case OS_NT4:
						if (!cl->m_DoNT4)
							goto skip;
						break;
					case OS_NT351: 
						if (!cl->m_DoNT351)
							goto skip;
						break;
				}	

				// We probarbly need these files. Link to global list
				cl->m_Next = m_FileClassRoot;
				m_FileClassRoot = cl;
			}
			else
				IsScanningClass = FALSE;

			continue;
		}
		if ((*ReadBuf == 1) && IsScanningClass)
		{
			fl = new CFileList;

			// Link on top (to make it easy)
			fl->m_Next = cl->m_FileList;
			cl->m_FileList = fl;

			MySscanf(ReadBuf,"\1%s\1 \1%s\1 %d \1%s\1",
				fl->m_SrcName,
				fl->m_DstName,
				&fl->m_MenuMode,
				fl->m_MenuName);

			if (!strcmp(fl->m_DstName,"(NULL)"))
				strcpy(fl->m_DstName, StripPath(fl->m_SrcName));
			else
			{
				if (fl->m_DstName[strlen(fl->m_DstName)-1] == '\\')
					strcat(fl->m_DstName,StripPath(fl->m_SrcName));
			}
		}
	}

	return TRUE;
}

void CSetupApp::ExpandMacros(CString& cPath)
{
	CString cBuf("");
	LPCSTR p = cPath;

	while(*p)
	{
		if (*p == '<')
		{
			if (!strnicmp(p,"<Install Dir>", 13))
			{
				p += 13;
				cBuf = DestPath;
				if ((cBuf[cBuf.GetLength() -1] != '\\') && (*p != '\\'))
					cBuf += '\\';
			}
			else if (!strnicmp(p,"<Windows Dir>", 13))
			{
				p += 13;
				cBuf = WinDir;
				if ((cBuf[cBuf.GetLength() -1] != '\\') && (*p != '\\'))
					cBuf += '\\';
			}
			else if (!strnicmp(p,"<Windows System Dir>", 20))
			{
				p += 20;
				cBuf = WinDir;
				if ((cBuf[cBuf.GetLength() -1] != '\\') && (*p != '\\'))
					cBuf += '\\';
			}

		}
		else
			cBuf += *p++;
	}

	cPath = cBuf;
}

BOOL CSetupApp::InstallFile(CClassList *cl, CFileList *fl)
{
	// Check if target exist.
	DWORD dwRval;
	static char szCurDir[MAX_PATH];
	static char szDestDir[MAX_PATH];
	static char szTmpFile[MAX_PATH];
	static char szCWD[MAX_PATH];
	UINT lpuCurDirLen = MAX_PATH;
	UINT lpuDestDirLen = MAX_PATH;
	UINT lpuTmpFileLen = MAX_PATH;
	CString cBuf, cPath;
	int NumFound = 0;
	DWORD dwAttr;

	*szCurDir = 0;


	GetCurrentDirectory(MAX_PATH,szCWD);

	CString cDestPath = cl->m_InstallDir;
	if ((cDestPath[cDestPath.GetLength() -1] != '\\')
		&& (*cl->m_InstallDir != '\\'))
	{
		cDestPath += '\\';
	}

	cDestPath += fl->m_DstName;

	ExpandMacros(cDestPath);
	if (!cDestPath.IsEmpty() && (cDestPath[cDestPath.GetLength() -1] == '\\'))
	{
		cDestPath += StripPath(fl->m_SrcName);
		strcpy(fl->m_DstName, cDestPath);	
	}

	// Prepere destrination path (if present)
	char *PrgPath = "";
	char *p = strrchr(cDestPath,'\\');
	if (p)
	{
		PrgPath = strdup(cDestPath);
		p = strrchr(PrgPath,'\\');
		*++p = 0;
	}

	m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"InstallFile: %s\n", PrgPath);

	// Make a quick scan for the file in Windows, system and system32
	if (cl->m_IsSystem)
	{
		// Prepere the error message in case we need it.
		cBuf.Format(
			"Warning. Multiple copies of the shared file %s is found in the following paths:",
			 StripPath(fl->m_DstName));

		// Try Windows
		cPath.Format("%s\\%s", WinDir, StripPath(fl->m_DstName));
		if (GetFilePermissions(cPath, dwAttr))
		{
			strcpy(szCurDir,WinDir);
			cBuf += "\r\n"; cBuf += cPath; NumFound++;
			m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\tTarget found in %s\n", WinDir);
		}

		// Try Windows\system
		cPath.Format("%s\\System\\%s", WinDir, StripPath(fl->m_DstName));
		if (GetFilePermissions(cPath,dwAttr))
		{
			sprintf(szCurDir,"%s\\system",WinDir);
			cBuf += "\r\n"; cBuf += cPath; NumFound++;
			m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\tTarget found in %s\\System\n", WinDir);
		}

		// Try Windows\system32
		cPath.Format("%s\\System32\\%s", WinDir, StripPath(fl->m_DstName));
		if (GetFilePermissions(cPath,dwAttr))
		{
			sprintf(szCurDir,"%s\\system32",WinDir);
			cBuf += "\r\n"; cBuf += cPath; NumFound++; 
			m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\tTarget found in %s\\System32\n", WinDir);
		}

		if (NumFound > 1)
		{
			cBuf += "\r\nThis can cause problems with this and other programs. Please consider appropriate actions to resolve this possible problem.";
			AfxMessageBox(cBuf);
			m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\tUser alerted about %d targets...\n", NumFound);
		}
	}

	if (strchr(PrgPath,':'))
	{
		cBuf = PrgPath;
		if (cBuf[cBuf.GetLength() -1] == '\\')
			cBuf.ReleaseBuffer(cBuf.GetLength() -1);
	}
	else
		cBuf = cl->m_IsSystem ? SystemDir : DestPath;

	if (*PrgPath && !strchr(PrgPath,':'))
	{
		if (cBuf[cBuf.GetLength() -1] != '\\')
			cBuf += '\\';
		cBuf += PrgPath;
		if (cBuf[cBuf.GetLength() -1] == '\\')
			cBuf.ReleaseBuffer(cBuf.GetLength() -1);
	}

	if (*PrgPath)
	{
		if (CreateDirectory(cBuf,NULL))
		{
			fprintf(m_UnInstFp,"dir %s\n", cBuf);
			m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\tDirectory %s created.\n", cBuf);
		}
	}

	{
		dwRval = 0;
		strcpy(szDestDir,cBuf);
		if (!*szCurDir)
			strcpy(szCurDir,cBuf);
	}

	cPath.Format("%s\\%s",szDestDir, StripPath(fl->m_DstName));

	// Try to install the file
	dwRval = MyVerInstallFile(
		0,
		(char *)StripPath(fl->m_SrcName),
		(char *)StripPath(fl->m_DstName),
		szCWD,
		szDestDir,
		szCurDir,
		szTmpFile,
		&lpuTmpFileLen);

	m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\tMyVerInstallFile() returned %d\n", dwRval);

	if (dwRval & VIF_SRCOLD)
	{
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\tTarget is newer than the install fille. Skipping...\n");
		return FALSE;
	}

	if (dwRval & VIF_OUTOFSPACE)
	{
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\tVIF_OUTOFSPACE. Skipping...\n");
		AfxMessageBox("The file can not be copied due to insufficient disk space.");
		return FALSE;
	}

	if (dwRval)
	{
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\tTrying to force installation...\n");
		// Try to force the installation!
		// VIF_OUTOFSPACE
		dwRval = MyVerInstallFile(
		VIFF_FORCEINSTALL,
		(char *)StripPath(fl->m_SrcName),
		(char *)StripPath(fl->m_DstName),
		szCWD,
		szDestDir,
		szCurDir,
		szTmpFile,
		&lpuTmpFileLen);
	}
	
	// Somehow it reports out of memory under NT 351 when the file is in use
	if (dwRval & (VIF_FILEINUSE | VIF_OUTOFMEMORY))
	{
		// File is in use. We have to schedule the install for the next system reboot.
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\tFile is in use... ");
		if ((OsType == OS_NT4) || (OsType == OS_NT351))
		{
			// Use NT MoveFileEx
			if (MoveFileEx(StripPath(fl->m_SrcName), cPath, MOVEFILE_DELAY_UNTIL_REBOOT|MOVEFILE_REPLACE_EXISTING|MOVEFILE_COPY_ALLOWED))
			{
				m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"NT: ***MoveFileEx() failed!\n");
				AfxMessageBox("MoveFileEx() failed.");
			}
			else
			{
				m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"NT: MoveFileEx() OK\n");
				m_DoReboot = TRUE;
			}
		}
		else if (OsType == OS_WIN95)
		{
			cBuf = cPath;
			cBuf += ".tmp";
			MoveFile(StripPath(fl->m_SrcName), cBuf);
			WritePrivateProfileString("rename", cPath, cBuf, "wininit.ini");
			m_DoReboot = TRUE;
			m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"Win95: Scheduled replace at boot OK\n");
		}
		else
		{
			m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"*** Dont know how to install shared files on this system. Shared files will not be installed. if they are in use by Windows.\n");
			AfxMessageBox("Dont know how to install shared files on this system. Shared files will not be installed. if they are in use by Windows.");
			return FALSE;
		}
	}
	else if (dwRval)
	{
		cBuf.Format("Failed to install file: %s\\%s --> %s\\%s\r\nFlags = ", 
		szCWD,
		StripPath(fl->m_SrcName),
		szDestDir,
		StripPath(fl->m_DstName));

		if (dwRval & VIF_TEMPFILE) cBuf += "VIF_TEMPFILE";
		if (dwRval & VIF_MISMATCH) cBuf += "VIF_MISMATCH";
		if (dwRval & VIF_SRCOLD) cBuf += "VIF_SRCOLD";
		if (dwRval & VIF_DIFFLANG) cBuf += "VIF_DIFFLANG";
		if (dwRval & VIF_DIFFCODEPG) cBuf += "VIF_DIFFCODEPG";
		if (dwRval & VIF_DIFFTYPE) cBuf += "VIF_DIFFTYPE";
		if (dwRval & VIF_WRITEPROT) cBuf += "VIF_WRITEPROT";
		if (dwRval & VIF_FILEINUSE) cBuf += "VIF_FILEINUSE";
		if (dwRval & VIF_OUTOFSPACE) cBuf += "VIF_OUTOFSPACE";
		if (dwRval & VIF_ACCESSVIOLATION) cBuf += "VIF_ACCESSVIOLATION";
		if (dwRval & VIF_SHARINGVIOLATION) cBuf += "VIF_SHARINGVIOLATION";
		if (dwRval & VIF_CANNOTCREATE) cBuf += "VIF_CANNOTCREATE";
		if (dwRval & VIF_CANNOTDELETECUR) cBuf += "VIF_CANNOTDELETECUR";
		if (dwRval & VIF_CANNOTRENAME) cBuf += "VIF_CANNOTRENAME";
		if (dwRval & VIF_OUTOFMEMORY) cBuf += "VIF_OUTOFMEMORY";
		if (dwRval & VIF_CANNOTREADSRC) cBuf += "VIF_CANNOTREADSRC";
		if (dwRval & VIF_BUFFTOOSMALL) cBuf += "VIF_BUFFTOOSMALL";
		if (dwRval & VIF_CANNOTREADDST) cBuf += "VIF_CANNOTREADDST";
		AfxMessageBox(cBuf);
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"%s\n", cBuf);
		return FALSE;
	}

	if (((OsType == OS_NT4) || (OsType == OS_NT351)) && ((cl->m_IsShared || cl->m_IsSystem)))
	{
		// Increment Regitry counters for shared files
		IncRegCnt(cPath);
	}

	m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\tFile is installed or scheduled OK\n");

	// Add log line
	fprintf(m_UnInstFp,"file %d %d %s\n", 
		cl->m_IsShared, cl->m_IsSystem, (LPCSTR)cPath);

	if (fl->m_MenuMode)
	{
		if ((OsType == OS_NT4) || (OsType == OS_WIN95))
		{
			cBuf = (fl->m_MenuMode == 1) ? UserProgramFolderDir : ProgramFolderDir;
			if (cBuf.IsEmpty())
			{
				static BOOL HaveTold;
				if (HaveTold)
				{
					m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"Warning: Setup can not create shortcut to the Start Menu. You have to add the application to the Start Menu manually...\n");
					AfxMessageBox("Warning: Setup can not create shortcut to the Start Menu. You have to add the application to the Start Menu manually...");
					HaveTold = TRUE;
				}
				
				goto skip;
			}
			cBuf += '\\';
			cBuf += FolderName;
			fprintf(m_UnInstFp,"dir %s\n", (LPCSTR)cBuf);
			CreateDirectory(cBuf,NULL);
			cBuf += '\\';
			cBuf += fl->m_MenuName;
			cBuf += ".lnk";
			
			CreateLink(cPath,cBuf,m_Name);
			fprintf(m_UnInstFp,"file 0 0 %s\n", (LPCSTR)cBuf);
		}
		else if (OsType == OS_NT351)
		{
			fprintf(m_UnInstFp,"pmngr %s@%s\n", (LPCSTR)FolderName, fl->m_MenuName);
			DDECreateGroup(FolderName);
			DDEAddProgramItem(fl->m_MenuName,cPath);
		}
	}
skip:

	return TRUE;
}

LPCSTR StripPath(LPCSTR Path)
{
	LPCSTR Name;

	if ((Name = strrchr(Path,'\\')) != NULL)
		++Name;
	else
		Name = Path;

	return Name;
}


// CreateLink - uses the shell's IShellLink and IPersistFile interfaces 
//   to create and store a shortcut to the specified object. 
// Returns the result of calling the member functions of the interfaces. 
// lpszPathObj - address of a buffer containing the path of the object 
// lpszPathLink - address of a buffer containing the path where the 
//   shell link is to be stored 
// lpszDesc - address of a buffer containing the description of the 
//   shell link 
 
HRESULT CreateLink(LPCSTR lpszPathObj, LPCSTR lpszPathLink, LPCSTR lpszDesc) 
{ 
    HRESULT hres; 
    IShellLink* psl; 
		LPSTR WorkingDir = strdup(lpszPathObj);
		LPSTR p = strrchr(WorkingDir,'\\');
		if (p && *p)
			*p = 0;
 
    // Get a pointer to the IShellLink interface. 
    hres = CoCreateInstance(CLSID_ShellLink, NULL, 
        CLSCTX_INPROC_SERVER, IID_IShellLink, (void **)&psl); 
    if (SUCCEEDED(hres)) { 
        IPersistFile* ppf; 
 
        // Set the path to the shortcut target, and add the 
        // description. 
        psl->SetPath(/*psl,*/ lpszPathObj); 
        psl->SetDescription(/*psl,*/ lpszDesc); 
				psl->SetWorkingDirectory(WorkingDir);
 
       // Query IShellLink for the IPersistFile interface for saving the 
       // shortcut in persistent storage. 
        hres = psl->QueryInterface(/*psl, */IID_IPersistFile, 
            (void **)&ppf); 
 
        if (SUCCEEDED(hres)) { 
            WORD wsz[MAX_PATH]; 
 
            // Ensure that the string is ANSI. 
            MultiByteToWideChar(CP_ACP, 0, lpszPathLink, -1, 
                wsz, MAX_PATH); 
 
            // Save the link by calling IPersistFile::Save. 
            hres = ppf->Save(/*ppf,*/ wsz, TRUE); 
            ppf->Release(/*ppf*/); 
        } 
        psl->Release(/*psl*/); 
    } 
    return hres; 
} 
 

BOOL CSetupApp::IncRegCnt(LPCSTR Path)
{
	DWORD dwNihl, dwSize = sizeof(DWORD);
	LONG Rval;
	HKEY hkey;
	DWORD dwCounter = 1;
	
	if ((Rval = RegOpenKeyEx(
		HKEY_LOCAL_MACHINE,
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\SharedDLLs",
		0,
		KEY_QUERY_VALUE | KEY_WRITE,
		&hkey)) == ERROR_SUCCESS) 
	{
		if ((Rval = RegQueryValueEx(
			hkey,
			Path,
			NULL,
			&dwNihl,
			(LPBYTE)&dwCounter,
			&dwSize)) == ERROR_SUCCESS)
		{
			++dwCounter;
		}
		if (RegSetValueEx(
			hkey,
			Path,
			0,
			REG_DWORD,
			(LPBYTE)&dwCounter,
			dwSize) != ERROR_SUCCESS)
		{
			AfxMessageBox("Failed to increment the registery counter for this shared file.");
		}

		RegCloseKey(hkey);
		fprintf(m_UnInstFp,"inc %s\n", Path);
		return TRUE;
	}
	else
	{
		CString cBuf;
		cBuf.Format("Failed to open the registery counter for this shared file. Error = %u",
			Rval);
		AfxMessageBox(cBuf);
	}

	return FALSE;
}

// Have to do this job by ourself since
// the library function dont work with 
// long filenames under Win95 and God knows what else...
DWORD CSetupApp::MyVerInstallFile(DWORD uFlags,LPTSTR szSrcFileName,
	LPTSTR szDestFileName,LPTSTR szSrcDir,LPTSTR szDestDir,
	LPTSTR szCurDir, LPTSTR szTmpFile,PUINT lpuTmpFileLen)
{

	m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t  MyVerInstallFile(%s --> %s)\n", szSrcFileName, szDestFileName);
	BOOL HaveBoth = FALSE;
	// We don't use temporary files :-)
	szTmpFile = "";
	*lpuTmpFileLen = 0;

	if (!strchr(szCurDir,'\\'))
		*szCurDir = 0;

	// Make the full paths
	CString cSrcPath, cDestPath;

	cSrcPath.Format("%s\\%s", szSrcDir, szSrcFileName);
	cDestPath.Format("%s\\%s", szCurDir, szDestFileName);

	m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tSrc = %s\n", cSrcPath);
	m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tDst = %s\n", cDestPath);

	// Start with a simple check to see if both files exist
	HANDLE hSrc, hDest;
	WIN32_FIND_DATA dSrc, dDest;
	
	if ((hSrc = FindFirstFile(cSrcPath, &dSrc)) == INVALID_HANDLE_VALUE)
	{
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\t**VIF_CANNOTREADSRC\n");
		return VIF_CANNOTREADSRC;
	}
	FindClose(hSrc);
	m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tSrc exist.\n");

	if (*szCurDir && !(uFlags & VIFF_FORCEINSTALL))
	{
		// We must check if the destination file is newer than the source file
		if ((hDest = FindFirstFile(cDestPath, &dDest)) != INVALID_HANDLE_VALUE)
		{
			// Both files exist.
			m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tDst exist\n");
			FindClose(hDest);
			HaveBoth = TRUE;

			// Compare verison
			DWORD BufferSize, dwDummy;
			UINT Dummy;
			char *Vbuf = NULL;
			LPVOID vPtr;
			VS_FIXEDFILEINFO vSrc, vDest;

			BufferSize = GetFileVersionInfoSize((char *)(LPCSTR)cSrcPath,&dwDummy);
			if (!BufferSize)
			{
				Dummy = GetLastError();
				m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tFailed to get version info size on src. Error %d\n", Dummy); 
				switch(Dummy)
				{
					case ERROR_FILE_NOT_FOUND:
					case ERROR_PATH_NOT_FOUND:
					case ERROR_ACCESS_DENIED:
					case ERROR_NO_MORE_FILES:
					case ERROR_TOO_MANY_OPEN_FILES:
					case ERROR_OUTOFMEMORY:
					case ERROR_SHARING_VIOLATION:
					case ERROR_NETWORK_BUSY:
						return VIF_CANNOTREADSRC;
					default:
						goto datecmp; // We assume that the file dont have a version
				}
			}
				
			Vbuf = new char[BufferSize];

			if (!GetFileVersionInfo((char *)(LPCSTR)cSrcPath,0,BufferSize,(LPVOID)Vbuf))
			{
				//return VIF_CANNOTREADSRC;
				m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tFailed to get version information on src.\n"); 
				goto datecmp;
			}

			if (!VerQueryValue(Vbuf,TEXT("\\"), &vPtr, &Dummy))
			{
				//return VIF_CANNOTREADSRC;
				m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tsrc: VerQueryValue() failed.\n"); 
				goto datecmp;
			}

			ASSERT(BufferSize >= sizeof(VS_FIXEDFILEINFO));
			ASSERT(Dummy >= sizeof(VS_FIXEDFILEINFO));

			memcpy(&vSrc,vPtr,sizeof(VS_FIXEDFILEINFO));
			delete Vbuf;

				// Get the destination info.
			BufferSize = GetFileVersionInfoSize((char *)(LPCSTR)cDestPath,&dwDummy);
		
			if (!BufferSize)
			{
				Dummy = GetLastError();
				m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tFailed to get version info size on dst. Error %d\n", Dummy); 
				switch(Dummy)
				{
					case ERROR_FILE_NOT_FOUND:
					case ERROR_PATH_NOT_FOUND:
					case ERROR_ACCESS_DENIED:
					case ERROR_NO_MORE_FILES:
					case ERROR_TOO_MANY_OPEN_FILES:
					case ERROR_OUTOFMEMORY:
					case ERROR_SHARING_VIOLATION:
					case ERROR_NETWORK_BUSY:
						return VIF_CANNOTREADDST;
					default:
						goto copy; // We assume that the file dont have a version
				}
			}
		
			Vbuf = new char[BufferSize];

			if (!GetFileVersionInfo((char *)(LPCSTR)cDestPath,0,BufferSize,(LPVOID)Vbuf))
			{
				//return VIF_CANNOTREADSRC;
				m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tFailed to get version information on src.\n"); 
				goto datecmp;
			}

			if (!VerQueryValue(Vbuf,TEXT("\\"), &vPtr, &Dummy))
			{
				m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tdst: VerQueryValue() failed.\n"); 
				//return VIF_CANNOTREADSRC;
				goto datecmp;
			}

			ASSERT(BufferSize >= sizeof(VS_FIXEDFILEINFO));
			ASSERT(Dummy >= sizeof(VS_FIXEDFILEINFO));
			memcpy(&vDest,vPtr,sizeof(VS_FIXEDFILEINFO));
	
			delete Vbuf;

			m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tComparsion are done by file version.\n");
			m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tsrc=(%u,%u) dst=(%u,%u).\n",
				vSrc.dwFileVersionMS, vSrc.dwFileVersionLS,
				vDest.dwFileVersionMS, vDest.dwFileVersionLS );
	
			// Lets compare the version numbers
			if ((vDest.dwFileVersionMS > vSrc.dwFileVersionMS) 
				|| (vDest.dwFileVersionMS == vSrc.dwFileVersionMS) && (vDest.dwFileVersionLS >= vSrc.dwFileVersionLS))
			{
				m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tDst are more up to date than src.\n");
				return VIF_SRCOLD | VIF_MISMATCH;
			}

			// If we get this far, we *ignore* the file date..
			goto copy;

datecmp:
			m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tComparsion are defaulting to file date...\n");
			// Compare file dates
			if (HaveBoth && CompareFileTime(&dSrc.ftLastWriteTime,&dDest.ftLastWriteTime) <= 0)
			{
				m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tDst are newer than the install file.\n");
				return VIF_SRCOLD;
			}

			// TODO: Add comparsion on language, OS and codepage.
		}
	}
copy:

	m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tPrepering for copy...\n");

	if (*szCurDir && !(uFlags & VIFF_DONTDELETEOLD))
	{
		// Delete the old file if it is read-only
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tOld file must (eventually) be deleted... ");
		DWORD dwAttr = 0;
		if (GetFilePermissions(cDestPath,dwAttr) && (dwAttr & FILE_ATTRIBUTE_READONLY))
		{
			if (!(uFlags & VIFF_FORCEINSTALL))
			{
				m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"Not called with permission to delete readonly files... Aborting operation.\n");
				return VIF_WRITEPROT;
			}
			
			CFileStatus fs;
			CFile::GetStatus(cDestPath,fs);
			fs.m_attribute &= ~CFile::readOnly;
			CFile::SetStatus(cDestPath,fs);
			m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"File is prepered for delete... ");
		}
		if (!DeleteFile(cDestPath))
		{
			DWORD Err = GetLastError();
			if (Err != ERROR_FILE_NOT_FOUND)
			{
				m_LogFp == NULL ? 0 :  fprintf(m_LogFp," **Failed. Error code %d\n", Err);
				if (dwAttr & FILE_ATTRIBUTE_SYSTEM)
				{
					m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tSystem attribute on file.. Assuming file in use...\n");
					return VIF_FILEINUSE;
				}

				m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\t**Aborting operation du to error.\n");
				switch(Err)
				{
					case ERROR_ACCESS_DENIED:
						return VIF_FILEINUSE | VIF_CANNOTDELETE;
				
					case ERROR_PATH_NOT_FOUND:
					case ERROR_NO_MORE_FILES:
					case ERROR_TOO_MANY_OPEN_FILES:
					case ERROR_OUTOFMEMORY:
					case ERROR_SHARING_VIOLATION:
					case ERROR_NETWORK_BUSY:
						return VIF_CANNOTDELETE;
				}
			}
		}
	}


	cDestPath.Format("%s\\%s", szDestDir, szDestFileName);
	// Lets do the transfere.

	if (!CopyFile(cSrcPath,cDestPath,FALSE))
	{
		CString cBuf;
		cBuf.Format("%s --> %s\r\nCopy failed.", cSrcPath, cDestPath);
		AfxMessageBox(cBuf);
		DWORD Err = GetLastError();
		m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tcBuf Error code %d\n", cBuf, Err);
		switch(Err)
		{
			case ERROR_FILE_NOT_FOUND:
				return VIF_CANNOTREADSRC;
			case ERROR_SHARING_VIOLATION:
				return VIF_SHARINGVIOLATION | VIF_FILEINUSE;
			case ERROR_ACCESS_DENIED:
				return VIF_ACCESSVIOLATION;
			default:
				return VIF_CANNOTCREATE;
		}
	}
	m_LogFp == NULL ? 0 :  fprintf(m_LogFp,"\t\tFile installed OK.\n");
	return 0;
}


BOOL GetFilePermissions(LPCSTR Path, DWORD& Perms)
{
	Perms = GetFileAttributes(Path);
	
	return Perms != 0xFFFFFFFF;
}


