// Uninstall.cpp

#include "windows.h"
#include <stdio.h>
#include <ctype.h>

#include "ddeml.h"

int Token(LPCSTR Line, LPCSTR *Tok);
BOOL DecrementRegKey(LPCSTR Path);
BOOL DeleteRegKey(LPCSTR DeleteKey);
void RemoveFolderIcon(LPCSTR FolderNameOrg);

extern BOOL DDEDeleteGroup(LPCSTR GroupName);
extern BOOL DDEIsFolderEmpty(LPCSTR FolderName);
extern BOOL DDEDeleteIcon(LPCSTR GroupName, LPCSTR IconName);

CONVCONTEXT   CCFilter = { sizeof (CONVCONTEXT), 0, 0, 0, 0L, 0L };
static DWORD lIdInst, lIdInst2;

HDDEDATA CALLBACK GroupDDECallback (UINT uiType, UINT uiFmt, HANDLE hConv,
      HSZ sz1, HSZ sz2, HDDEDATA hData, LONG lData1, LONG lData2) 
{
	return ((HDDEDATA) NULL);
}/* end GroupDDECallback */


enum // Tokens
{
	T_INC,
	T_FILE,
	T_KEY,
	T_DIR,
	T_PICON
};

typedef struct LinesHdr
{
	char *Buf;
	struct LinesHdr *Next;
} Lines;

Lines  *LinesRoot = NULL;
char *UninstFile = ".UnInst.inf";

int main(int argc, char **argv)
{
	FILE *fp;
	char buf[MAX_PATH*4], *p;
	LPCSTR Tok;
	Lines *ll;

	while(--argc)
	{
		p = *++argv;

		if (*p == '/')
			printf("Unknown argument %s\n", p);
		else
			UninstFile = p;
	}

	if ((fp = fopen(UninstFile,"r")) == NULL)
	{
		printf("Unable to open %s\nPress ENTER to exit.", UninstFile);
		getc(stdin);
		exit(1);
	}

	*buf = 0;
	fgets(buf,sizeof(buf),fp);

	printf("This program will uninstall %s\n\nPress Y and [ENTER] to uninstall or just [ENTER] to abort.",
		buf);

	if (toupper(getc(stdin)) != 'Y')
		exit(2);

	printf("%s will now be uninstalled\n\n", buf);

	// Build list in memory. We must process the .inf file in reverse order
	while(fgets(buf,sizeof(buf),fp) != NULL)
	{
		ll = new Lines;
		ll->Buf = strdup(buf);
		ll->Next = LinesRoot;
		LinesRoot = ll;
		if ((p = strchr(ll->Buf,'\n')) != NULL)
			*p = 0;
	}

	fclose(fp);

	for(ll = LinesRoot; ll; ll = ll->Next)
	{
		switch(Token(ll->Buf,&Tok))
		{
			case T_INC:
				// Decrement counter
				DecrementRegKey(Tok);
				break;

			case T_FILE:
				if ((*Tok == '0') && (Tok[2] == '0'))
				{
					// Not a shared file. 
					// Shared files are deleted by the decrement function...
					printf("Deleting file %s ", Tok + 4);
					if (DeleteFile(Tok + 4))
						printf(" .. Deleted.\n");
					else
						printf(" .. Failed to delete file.\n");
				}

			case T_KEY:
				DeleteRegKey(Tok);
				break;

			case T_DIR:
				RemoveDirectory(Tok);
				break;

			case T_PICON:
				RemoveFolderIcon(Tok);
				break;
		}
	}

	DeleteFile(UninstFile);

	printf("Done. Press [ENTER] to exit.");
	fflush(stdin);
	getc(stdin);
	
	return 0;
}


int Token(LPCSTR Line, LPCSTR *Tok)
{
	switch(*Line)
	{
		case 'i': // Inc
			*Tok = Line + 4;
			return T_INC;
		case 'f': // File
			*Tok = Line + 5;
			return T_FILE;
		case 'k': // Registery key
			*Tok = Line + 4;
			return T_KEY;
		case 'd': // Directory
			*Tok = Line + 4;
			return T_DIR;
		case 'p': // Program manager folder icon (Nt 3.51)
			*Tok = Line + 6;
			return T_PICON;
	}
	return -1;
}

BOOL DecrementRegKey(LPCSTR Path)
{
	DWORD dwNihl, dwSize = sizeof(DWORD);
	LONG Rval;
	HKEY hkey;
	DWORD dwCounter = 0;
	
	if (RegOpenKeyEx(
		HKEY_LOCAL_MACHINE,
		"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\SharedDLLs\\",
		0,
		KEY_QUERY_VALUE | KEY_WRITE,
		&hkey) == ERROR_SUCCESS) 
	{
		if ((Rval = RegQueryValueEx(
			hkey,
			Path,
			NULL,
			&dwNihl,
			(LPBYTE)&dwCounter,
			&dwSize)) == ERROR_SUCCESS)
		{
			if (dwCounter > 0)
			{
					--dwCounter;	

				if (dwCounter == 0)
				{
					if (DeleteFile(Path))
					{
						char *MyBuf = new char[MAX_PATH * 3];
						sprintf(
							"HKEY_LOCAL_MACHINE SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\SharedDLLs\\%s",
							Path);
						RegCloseKey(hkey);
						DeleteRegKey(MyBuf);
						delete MyBuf;
						return TRUE;
					}
				}
				
				RegSetValueEx(
					hkey,
					Path,
					0,
					REG_DWORD,
					(LPBYTE)&dwCounter,
					dwSize);
				
			}
		}

		RegCloseKey(hkey);
		return TRUE;
	}
	return FALSE;
}


BOOL DeleteRegKey(LPCSTR DeleteKey)
{
	HKEY hkey, MainKey;
	char szKey[MAX_PATH];
	LPSTR Tag;
	char *MyBuf = strdup(DeleteKey);
	char *p = strtok(MyBuf," \n");

	if (!strcmp(p,"HKEY_CURRENT_USER"))
		MainKey = HKEY_CURRENT_USER;
	else if (!strcmp(p,"HKEY_CLASSES_ROOT"))
		MainKey = HKEY_CLASSES_ROOT;
	else if (!strcmp(p,"HKEY_LOCAL_MACHINE"))
		MainKey = HKEY_LOCAL_MACHINE;
	else if (!strcmp(p,"HKEY_USERS"))
		MainKey = HKEY_USERS;
	else
		return FALSE;

	p = strtok(NULL,"\n");

	strcpy(szKey, p);

	delete MyBuf;

	Tag = strrchr(szKey,'\\');
	*++Tag = 0;

	Tag = strrchr(DeleteKey,'\\');
	++Tag;

	if (RegOpenKeyEx(
		MainKey,
		szKey,
		0,
		KEY_WRITE,
		&hkey) == ERROR_SUCCESS) 
	{
		RegDeleteKey(hkey,Tag);
		RegCloseKey(hkey);
	}

	return FALSE;
}


void RemoveFolderIcon(LPCSTR FolderNameOrg)
{
	LPSTR FolderName = strdup(FolderNameOrg);

	char *IconName = strchr(FolderName,'@');

	if (!IconName || !*IconName)
		return;

	*IconName++ = 0;

	DDEDeleteIcon(FolderName, IconName);

	if (DDEIsFolderEmpty(FolderName))
		DDEDeleteGroup(FolderName);

	delete FolderName;
}

BOOL DDEIsFolderEmpty(LPCSTR FolderName)
{
	LPBYTE     lpByte;
	LPSTR      szGroups;
	LPSTR      szToken;
	LPSTR      szMem;
	DWORD       lResult;
	DWORD       lIdLocal;
	HCONV      hConv;
	HDDEDATA   hDdeData;
	HSZ        szProgMan;
	HSZ        szTopic;
	BOOL		   Rval = FALSE;


	// Initialize the Dde id to 0

	lIdLocal = 0L;
	szGroups = NULL;

	// Attempt to initialize a conversation.

	if (DdeInitialize (&lIdLocal, (PFNCALLBACK) GroupDDECallback,
		(DWORD) APPCMD_CLIENTONLY, 0L)) 
	{ 
		return (FALSE);
	}/*endIf*/

	// Start a critical section. This fixes a problem where the DDEML
	// Can hang under threaded conditions.


	// Create a string handle for the Dde conversation.

	szProgMan = DdeCreateStringHandle (lIdLocal, "PROGMAN", CP_WINANSI);
	
	// Connect to the program manager.

	hConv = DdeConnect (lIdLocal, szProgMan, szProgMan, &CCFilter);

	// Create a handle for the Group topic.

	szTopic = DdeCreateStringHandle (lIdLocal, FolderName, CP_WINANSI);

	// Execute a request for the group names.

	hDdeData = DdeClientTransaction (NULL, 0L, hConv, szTopic, CF_TEXT,
		XTYP_REQUEST, 1000, &lResult);

	DdeFreeStringHandle (lIdLocal, szProgMan);
	DdeFreeStringHandle (lIdLocal, szTopic);

	// Disconnect the conversation.

	DdeDisconnect (hConv);

	// Verify that we got something back.

	if (hDdeData && (hDdeData != (HDDEDATA) TRUE)) 
	{

		// Access the data to get a character pointer.

		lpByte = DdeAccessData (hDdeData, &lResult);

		// allocate a block of memory to use.

		szGroups = szMem = (LPSTR)VirtualAlloc (NULL, lResult, MEM_COMMIT,
			PAGE_READWRITE);
		if (!szMem) 
		{
			lResult = GetLastError ();
		} 
		else 
		{

			// Copy the Dde data to our own block.

			memcpy (szGroups, lpByte, lResult);
		}/*endIf*/

		// Free the Dde data that is there. This is our responsibility.

		DdeUnaccessData (hDdeData);
		DdeFreeDataHandle (hDdeData);
	}/*endIf*/

	// If we have a memory block then parse it.

	if (szMem != NULL) 
	{

		// Search for 'cr'.

		// While we have a token use it.
		if ((szToken = strtok (szGroups, ",")) != NULL)
			if ((szToken = strtok (NULL, ",")) != NULL)
				if ((szToken = strtok (NULL, ",")) != NULL)
					Rval = atoi(szToken) == 0;

		// Release the memory block.

		VirtualFree (szMem, lResult, MEM_DECOMMIT);
	}/*endIf*/

	// Release the DDEML.

	DdeUninitialize (lIdLocal);

	return Rval;
}


BOOL DDEDeleteGroup(LPCSTR GroupName)
{
	HDDEDATA   hData;
	//LPSTR      szText;
	LPSTR      szCommand;
	HCONV      hConv;
	HSZ        szProgMan;
	DWORD       lResult;

	// Allocate a block of memory for the group name that the user has selected.
	//szText = VirtualAlloc (NULL, 64, MEM_COMMIT, PAGE_READWRITE);

	// Allocate a block of memory for the command to send to the Program Manager.
	szCommand = (LPSTR) VirtualAlloc (NULL, 128, MEM_COMMIT, PAGE_READWRITE);
	if (szCommand) 
	{
		// We got the memory so we can proceed.

		// Initialize the DDE conversation here.

		if (DdeInitialize (&lIdInst, (PFNCALLBACK) GroupDDECallback,
			(DWORD) APPCMD_CLIENTONLY, 0L)) 
		{
			printf("***DDEML Initialization Failure\n");
			//VirtualFree (szText, 128, MEM_DECOMMIT);
			VirtualFree (szCommand, 128, MEM_DECOMMIT);
			return FALSE;
		}/*endIf*/

		// Create a DDEML string handle for the Program Manager.
		szProgMan = DdeCreateStringHandle (lIdInst, "PROGMAN", CP_WINANSI);
		if (szProgMan) 
		{

			// String handle created so continue.

			// Establish a conversation with the Program Manager.
			hConv = DdeConnect (lIdInst, szProgMan, szProgMan, &CCFilter);

			// Create the group by creating the command line.
			// The program manager will activate an existing group if the name matches
			// a DDE attempt to create a new group so the create versus show is somewhat
			// redundant.
			sprintf (szCommand, "[DeleteGroup(%s)]", GroupName);

			// Create a DDEML data handle for the command.
			hData = DdeCreateDataHandle (lIdInst, (LPBYTE)szCommand,
				strlen (szCommand) + 1, 0, (HSZ) NULL, CF_TEXT, 0L);

			// Send the transaction to the server waiting a maximum of 10 seconds.
			// The server will release the data handle.
			if (!DdeClientTransaction ((LPBYTE) hData, 0xFFFFFFFF, hConv,
				(HSZ) NULL, 0, XTYP_EXECUTE, 10000, &lResult)) 
			{

				// If it fails get the error code.
				lResult = DdeGetLastError (lIdInst);
				printf("***DdeClientTransaction Failed");
			}/*endIf*/

						// Group already exists so activate it.
			sprintf (szCommand, "[ShowGroup(%s,1)]", GroupName);

			// Create a DDEML data handle for the command.
			hData = DdeCreateDataHandle (lIdInst, (LPBYTE)szCommand, strlen (szCommand),
				0, (HSZ) NULL, CF_TEXT, 0L);

			// Send the transaction to the server waiting a maximum of 10 seconds.
			// The server will release the data handle.
			if (!DdeClientTransaction ((LPBYTE) hData, 0xFFFFFFFF, hConv,
				(HSZ) NULL, 0, XTYP_EXECUTE, 10000, &lResult)) 
			{

				// If it fails get the error code.
				lResult = DdeGetLastError (lIdInst);
			}/*endIf*/
		
			// Release the Program Manager string handle.
			DdeFreeStringHandle (lIdInst, szProgMan);

			// Disconnect from the server.
			DdeDisconnect (hConv);
		} 
		else 
		{
			lResult = DdeGetLastError (lIdInst);
		}/*endIf*/

		// Free the two blocks of memory that were allocated.
		//VirtualFree (szText, 64, MEM_DECOMMIT);
		VirtualFree (szCommand, 64, MEM_DECOMMIT);

		// Uninitialize the conversation here so that resources are freed.

		DdeUninitialize (lIdInst);
		lIdInst = 0L;
		return (TRUE);
	} 
	else 
	{
		printf("****Memory Allocation failure");
	}
	
	return (FALSE);
}


BOOL DDEDeleteIcon(LPCSTR GroupName, LPCSTR IconName)
{
	HDDEDATA   hData;
	//LPSTR      szText;
	LPSTR      szCommand;
	HCONV      hConv = (HCONV)0;
	HSZ        szProgMan;
	DWORD       lResult;

	// Allocate a block of memory for the group name that the user has selected.
	//szText = VirtualAlloc (NULL, 64, MEM_COMMIT, PAGE_READWRITE);

	// Allocate a block of memory for the command to send to the Program Manager.
	szCommand = (LPSTR) VirtualAlloc (NULL, 128, MEM_COMMIT, PAGE_READWRITE);
	if (szCommand) 
	{
		// We got the memory so we can proceed.

		// Initialize the DDE conversation here.

		if (DdeInitialize (&lIdInst, (PFNCALLBACK) GroupDDECallback,
			(DWORD) APPCMD_CLIENTONLY, 0L)) 
		{
			printf("****DDEML Initialization Failure");
			//VirtualFree (szText, 128, MEM_DECOMMIT);
			VirtualFree (szCommand, 128, MEM_DECOMMIT);
			return FALSE;
		}/*endIf*/

		// Create a DDEML string handle for the Program Manager.
		szProgMan = DdeCreateStringHandle (lIdInst, "PROGMAN", CP_WINANSI);
		if (szProgMan) 
		{

			sprintf (szCommand, "[ShowGroup(%s,1)]", GroupName);

			// Create a DDEML data handle for the command.
			hData = DdeCreateDataHandle (lIdInst, (LPBYTE)szCommand, strlen (szCommand),
				0, (HSZ) NULL, CF_TEXT, 0L);

			// Send the transaction to the server waiting a maximum of 10 seconds.
			// The server will release the data handle.
			if (!DdeClientTransaction ((LPBYTE) hData, 0xFFFFFFFF, hConv,
				(HSZ) NULL, 0, XTYP_EXECUTE, 10000, &lResult)) 
			{

				// If it fails get the error code.
				lResult = DdeGetLastError (lIdInst);
			}/*endIf*/

			// Establish a conversation with the Program Manager.
			hConv = DdeConnect (lIdInst, szProgMan, szProgMan, &CCFilter);

			// Create the group by creating the command line.
			// The program manager will activate an existing group if the name matches
			// a DDE attempt to create a new group so the create versus show is somewhat
			// redundant.
			sprintf (szCommand, "[DeleteItem(%s)]", IconName);

			// Create a DDEML data handle for the command.
			hData = DdeCreateDataHandle (lIdInst, (LPBYTE)szCommand,
				strlen (szCommand) + 1, 0, (HSZ) NULL, CF_TEXT, 0L);

			// Send the transaction to the server waiting a maximum of 10 seconds.
			// The server will release the data handle.
			if (!DdeClientTransaction ((LPBYTE) hData, 0xFFFFFFFF, hConv,
				(HSZ) NULL, 0, XTYP_EXECUTE, 10000, &lResult)) 
			{

				// If it fails get the error code.
				lResult = DdeGetLastError (lIdInst);
				printf("****DdeClientTransaction Failed");
			}/*endIf*/
		
			// Release the Program Manager string handle.
			DdeFreeStringHandle (lIdInst, szProgMan);

			// Disconnect from the server.
			DdeDisconnect (hConv);
		} 
		else 
		{
			lResult = DdeGetLastError (lIdInst);
		}/*endIf*/

		// Free the two blocks of memory that were allocated.
		//VirtualFree (szText, 64, MEM_DECOMMIT);
		VirtualFree (szCommand, 64, MEM_DECOMMIT);

		// Uninitialize the conversation here so that resources are freed.

		DdeUninitialize (lIdInst);
		lIdInst = 0L;
		return (TRUE);
	} 
	else 
	{
		printf("****Memory Allocation failure");
	}
	
	return (FALSE);
}


